package ink.wgink.module.activiti.listener.task;

import ink.wgink.exceptions.OaNodeException;
import ink.wgink.exceptions.SearchException;
import ink.wgink.interfaces.department.IDepartmentBaseService;
import ink.wgink.interfaces.department.IDepartmentUserBaseService;
import ink.wgink.interfaces.position.IPositionBaseService;
import ink.wgink.interfaces.position.IPositionUserBaseService;
import ink.wgink.interfaces.role.IRoleBaseService;
import ink.wgink.interfaces.role.IRoleUserBaseService;
import ink.wgink.interfaces.user.IUserBaseService;
import ink.wgink.module.activiti.enums.oa.assignee.*;
import ink.wgink.module.activiti.pojo.pos.oa.OaNodeAssigneePO;
import ink.wgink.module.activiti.service.activiti.IActivitiModelService;
import ink.wgink.module.activiti.service.oa.INodeAssigneeService;
import ink.wgink.module.activiti.service.oa.IOaFormReportService;
import ink.wgink.pojo.dtos.department.DepartmentUserDTO;
import ink.wgink.pojo.dtos.position.PositionDTO;
import ink.wgink.pojo.dtos.user.UserDTO;
import ink.wgink.util.ArrayListUtil;
import org.activiti.bpmn.model.UserTask;
import org.activiti.engine.RepositoryService;
import org.activiti.engine.RuntimeService;
import org.activiti.engine.delegate.DelegateExecution;
import org.activiti.engine.delegate.DelegateTask;
import org.activiti.engine.delegate.JavaDelegate;
import org.activiti.engine.delegate.TaskListener;
import org.activiti.engine.history.HistoricTaskInstance;
import org.activiti.engine.repository.ProcessDefinition;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

/**
 * @ClassName: OaUserTaskCompleteListener
 * @Description: OA用户任务节点完成监听器
 * @Author: wanggeng
 * @Date: 2022/3/28 10:31
 * @Version: 1.0
 */
@Component
public class OaUserTaskCompleteListener implements TaskListener, JavaDelegate {

    private static final Logger LOG = LoggerFactory.getLogger(OaUserTaskCompleteListener.class);
    @Autowired
    private RuntimeService runtimeService;
    @Autowired
    private RepositoryService repositoryService;
    @Autowired
    private IActivitiModelService activitiModelService;
    @Autowired
    private INodeAssigneeService nodeAssigneeService;
    @Autowired
    private IOaFormReportService oaFormReportService;
    @Autowired
    private IUserBaseService userBaseService;
    @Autowired
    private IDepartmentBaseService departmentBaseService;
    @Autowired
    private IDepartmentUserBaseService departmentUserBaseService;
    @Autowired
    private IRoleBaseService roleBaseService;
    @Autowired
    private IRoleUserBaseService roleUserBaseService;
    @Autowired
    private IPositionBaseService positionBaseService;
    @Autowired
    private IPositionUserBaseService positionUserBaseService;

    @Override
    public void notify(DelegateTask delegateTask) {
        LOG.debug(">>>> userTask complete");
        String taskDefinitionKey = delegateTask.getTaskDefinitionKey();
        String processDefinitionId = delegateTask.getProcessDefinitionId();
        LOG.debug("查找流程定义");
        ProcessDefinition processDefinition = repositoryService.getProcessDefinition(processDefinitionId);
        String deploymentId = processDefinition.getDeploymentId();

        Map<String, Object> variables = delegateTask.getVariables();
        LOG.debug("查询表单数据：variables: {}", variables);
        LOG.debug("获取下一节点");
        UserTask nextUserTask = activitiModelService.getNextUserTask(delegateTask.getTaskDefinitionKey(), delegateTask.getProcessDefinitionId(), variables);
        if (nextUserTask == null) {
            LOG.debug("下一个节点不存在");
            return;
        }
        LOG.debug("获取节点代理人配置");
        OaNodeAssigneePO oaNodeAssigneePO = nodeAssigneeService.getPOByDeploymentIdAndNodeId(deploymentId, taskDefinitionKey);
        if (oaNodeAssigneePO == null) {
            throw new OaNodeException("下一节点没有配置代理人信息");
        }
        if (StringUtils.equals(oaNodeAssigneePO.getNodeType(), NodeTypeEnum.NORMAL.getValue())) {
            LOG.debug("nodeType -> normal");
            normal(delegateTask, nextUserTask, oaNodeAssigneePO);
            return;
        }
        if (StringUtils.equals(oaNodeAssigneePO.getNodeType(), NodeTypeEnum.MULTIPLE.getValue())) {
            LOG.debug("nodeType -> multiple");
            multiple(delegateTask, nextUserTask, oaNodeAssigneePO);
            return;
        }

        if (nextUserTask != null) {
            if (StringUtils.equals(nextUserTask.getName(), "任务3")) {
//                nextUserTask.setAssignee("${signUser}");
//                MultiInstanceLoopCharacteristics characteristics = new MultiInstanceLoopCharacteristics();
//                characteristics.setInputDataItem("signUsers");
//                characteristics.setElementVariable("signUser");
//                characteristics.setSequential(false);
//                characteristics.setCompletionCondition("${nrOfCompletedInstances==nrOfInstances}");
//                nextUserTask.setLoopCharacteristics(characteristics);

                List<String> signUsers = new ArrayList<>();
                signUsers.add("1");
                signUsers.add("e48e9c4a-995e-4061-abcd-a3c260c11333");

//                runtimeService.setVariable(delegateTask.getExecutionId(), "signUsers", signUsers);
                delegateTask.setVariable("signUsers", signUsers);
            } else {
                nextUserTask.setAssignee("1");
            }
        }

    }

    @Override
    public void execute(DelegateExecution execution) {

    }

    /**
     * @param userTask
     * @param oaNodeAssigneePO
     */
    private void normal(DelegateTask delegateTask, UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {
        if (StringUtils.equals(oaNodeAssigneePO.getAssigneeType(), AssigneeTypeEnum.APPOINT.getValue())) {
            LOG.debug("assigneeType -> appoint");
            appoint(delegateTask, userTask, oaNodeAssigneePO);
            return;
        }
        if (StringUtils.equals(oaNodeAssigneePO.getAssigneeType(), AssigneeTypeEnum.AUTO.getValue())) {
            LOG.debug("assigneeType -> auto");
            auto(delegateTask, userTask, oaNodeAssigneePO);
            return;
        }
    }

    /**
     * 设置多实例节点
     *
     * @param delegateTask
     * @param userTask
     * @param oaNodeAssigneePO
     */
    private void multiple(DelegateTask delegateTask, UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {

    }

    /**
     * 手动指定
     *
     * @param delegateTask
     * @param userTask
     * @param oaNodeAssigneePO
     */
    private void appoint(DelegateTask delegateTask, UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {
        if (StringUtils.equals(oaNodeAssigneePO.getAssigneeCount(), AssigneeCountEnum.SINGLE.getValue())) {
            LOG.debug("assigneeCount -> single");
            String assignee = oaNodeAssigneePO.getAssignee();
            if (StringUtils.isBlank(assignee)) {
                throw new SearchException("下节点代理人未设置");
            }
            UserDTO userDTO = userBaseService.get(assignee);
            if (userDTO == null) {
                throw new SearchException("下节点代理人不存在");
            }
            LOG.debug("设置代理人");
            userTask.setAssignee(userDTO.getUserId());
            return;
        }
        if (StringUtils.equals(oaNodeAssigneePO.getAssigneeCount(), AssigneeCountEnum.CANDIDATE.getValue())) {
            LOG.debug("assigneeCount -> candidate");
            String candidates = oaNodeAssigneePO.getCandidates();
            if (StringUtils.isBlank(candidates)) {
                throw new SearchException("下节点候选人未设置");
            }
            List<UserDTO> userDTOs = userBaseService.listByUserIds(Arrays.asList(candidates.split(",")));
            if (userDTOs.isEmpty()) {
                throw new SearchException("下节点候选人不存在");
            }
            LOG.debug("设置候选人");
            userTask.setCandidateUsers(ArrayListUtil.listBeanStringIdValue(userDTOs, "userId", UserDTO.class));
            return;
        }
    }

    /**
     * 自动获取
     *
     * @param delegateTask
     * @param userTask
     * @param oaNodeAssigneePO
     */
    private void auto(DelegateTask delegateTask, UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {
        if (StringUtils.equals(oaNodeAssigneePO.getAutoAssignType(), AutoAssignTypeEnum.SCOPE.getValue())) {
            LOG.debug("autoAssigneeType -> scope");
            departmentType(delegateTask, userTask, oaNodeAssigneePO);
            return;
        }
        if (StringUtils.equals(oaNodeAssigneePO.getAutoAssignType(), AutoAssignTypeEnum.QUICK.getValue())) {
            LOG.debug("autoAssigneeType -> quick");
            quickAssignee(delegateTask, userTask, oaNodeAssigneePO);
            return;
        }
    }

    /**
     * 快捷指定
     *
     * @param delegateTask
     * @param userTask
     * @param oaNodeAssigneePO
     */
    private void quickAssignee(DelegateTask delegateTask, UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {
        String quickAssignee = oaNodeAssigneePO.getQuickAssignee();
        if (StringUtils.equals(quickAssignee, QuickAssigneeEnum.STARTER.getValue())) {
            LOG.debug("quickAssignee -> starter:发起人");
            HistoricTaskInstance startTask = activitiModelService.getStartTask(delegateTask.getProcessInstanceId());
            userTask.setAssignee(startTask.getAssignee());
            return;
        }
        if (StringUtils.equals(quickAssignee, QuickAssigneeEnum.STARTER_SUPERIOR_POSITION.getValue())) {
            LOG.debug("quickAssignee -> starterSuperiorPosition:发起人上级岗位");
            HistoricTaskInstance startTask = activitiModelService.getStartTask(delegateTask.getProcessInstanceId());
            String starterUserId = startTask.getAssignee();
            List<String> positionIds = positionUserBaseService.listPositionIdByUserId(starterUserId);
            List<PositionDTO> parentPositionDTOs = positionBaseService.listParentByPositionIds(positionIds);
            if (parentPositionDTOs.isEmpty()) {
                throw new OaNodeException("发起人不存在上级岗位");
            }
            List<String> parentPositionIds = ArrayListUtil.listBeanStringIdValue(parentPositionDTOs, "positionId", PositionDTO.class);
            LOG.debug("查询发起人上级岗位用户ID列表");
            List<String> parentPositionUserIds = positionUserBaseService.listUserId(parentPositionIds);
            LOG.debug("查询发起人上级岗位用户列表");
            List<UserDTO> parentPositionUserDTOs = userBaseService.listByUserIds(parentPositionUserIds);
            if (parentPositionUserDTOs.isEmpty()) {
                throw new OaNodeException("发起人上级岗位没有用户");
            }
            // 上级岗位可以有自己
            List<String> existParentPositionUserIds = ArrayListUtil.listBeanStringIdValue(parentPositionUserDTOs, "userId", UserDTO.class);
            LOG.debug("候选人用户列表");
            userTask.setCandidateUsers(existParentPositionUserIds);
            return;
        }
        if (StringUtils.equals(quickAssignee, QuickAssigneeEnum.PREVIOUS_STEP_ASSIGNEE_SUBORDINATE_POSITION.getValue())) {
            LOG.debug("quickAssignee -> previousStepAssigneeSubordinatePosition:上一步代理人下级岗位");
            return;
        }
        if (StringUtils.equals(quickAssignee, QuickAssigneeEnum.PREVIOUS_STEP_ASSIGNEE_SUPERIOR_POSITION.getValue())) {
            LOG.debug("quickAssignee -> previousStepAssigneeSuperiorPosition:上一步代理人上级岗位");
            return;
        }
        if (StringUtils.equals(quickAssignee, QuickAssigneeEnum.PREVIOUS_STEP_ASSIGNEE_SUPERIOR_POSITION.getValue())) {
            LOG.debug("quickAssignee -> previousStepAssigneeSameDepartment:上一步代理人同部门");
            return;
        }
        if (StringUtils.equals(quickAssignee, QuickAssigneeEnum.PREVIOUS_N_STEP_ASSIGNEE_SUBORDINATE_POSITION.getValue())) {
            LOG.debug("quickAssignee -> previousNStepAssigneeSubordinatePosition:前N步代理人");
            Integer previousNStep = oaNodeAssigneePO.getPreviousNStep();
            return;
        }
    }

    /**
     * 部门类型
     *
     * @param delegateTask
     * @param userTask
     * @param oaNodeAssigneePO
     */
    private void departmentType(DelegateTask delegateTask, UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {
        if (StringUtils.equals(oaNodeAssigneePO.getDepartmentType(), DepartmentTypeEnum.APPOINT.getValue())) {
            LOG.debug("departmentType -> appoint");
            String departmentIds = oaNodeAssigneePO.getDepartments();
            String roleIds = oaNodeAssigneePO.getRoles();
            String positionIds = oaNodeAssigneePO.getPositions();
            if (StringUtils.isBlank(departmentIds) && StringUtils.isBlank(roleIds) && StringUtils.isBlank(positionIds)) {
                throw new SearchException("下节点没有设置部门、角色、职位");
            }
            if (!StringUtils.isBlank(departmentIds)) {
                LOG.debug("查询部门用户");
                List<String> departmentUserIds = departmentUserBaseService.listUserId(Arrays.asList(departmentIds.split(",")));
                setDepartmentCandidateUser(delegateTask, userTask, departmentUserIds, roleIds, positionIds);
                return;
            }
            if (!StringUtils.isBlank(roleIds)) {
                LOG.debug("查询角色用户");
                List<String> roleUserIds = roleUserBaseService.listUserId(Arrays.asList(roleIds.split(",")));
                setRoleCandidateUser(delegateTask, userTask, roleUserIds, positionIds);
                return;
            }
            LOG.debug("查询职位用户");
            List<String> positionUserIds = positionUserBaseService.listUserId(Arrays.asList(positionIds.split(",")));
            setPositionCandidateUser(delegateTask, userTask, positionUserIds);
            return;
        }
        if (StringUtils.equals(oaNodeAssigneePO.getDepartmentType(), DepartmentTypeEnum.BELONG.getValue())) {
            LOG.debug("departmentType -> belong");
            String roleIds = oaNodeAssigneePO.getRoles();
            String positionIds = oaNodeAssigneePO.getPositions();
            List<DepartmentUserDTO> departmentUserDTOs = departmentUserBaseService.listByUserIds(Arrays.asList(delegateTask.getAssignee()));
            List<String> departmentIds = ArrayListUtil.listBeanStringIdValue(departmentUserDTOs, "departmentId", DepartmentUserDTO.class);
            if (departmentIds.isEmpty()) {
                throw new OaNodeException("代理人不存在部门");
            }
            List<String> departmentUserIds = departmentUserBaseService.listUserId(departmentIds);
            setDepartmentCandidateUser(delegateTask, userTask, departmentUserIds, roleIds, positionIds);
        }
    }

    /**
     * 设置部门候选用户
     *
     * @param delegateTask
     * @param userTask
     * @param departmentUserIds
     * @param roleIds
     * @param positionIds
     */
    private void setDepartmentCandidateUser(DelegateTask delegateTask, UserTask userTask, List<String> departmentUserIds, String roleIds, String positionIds) {
        if (!StringUtils.isBlank(roleIds)) {
            LOG.debug("查询角色用户");
            List<String> roleUserIds = roleUserBaseService.listUserIdByRoleIdsAndUserIds(Arrays.asList(roleIds.split(",")), departmentUserIds);
            setRoleCandidateUser(delegateTask, userTask, roleUserIds, positionIds);
            return;
        }
        setCandidateUsers(delegateTask, userTask, departmentUserIds);
    }

    /**
     * 设置角色候选用户
     *
     * @param delegateTask
     * @param userTask
     * @param roleUserIds
     * @param positionIds
     */
    private void setRoleCandidateUser(DelegateTask delegateTask, UserTask userTask, List<String> roleUserIds, String positionIds) {
        if (!StringUtils.isBlank(positionIds)) {
            LOG.debug("查询职位用户");
            List<String> positionUserIds = positionUserBaseService.listUserIdByPositionIdsAndUserIds(Arrays.asList(positionIds.split(",")), roleUserIds);
            setPositionCandidateUser(delegateTask, userTask, positionUserIds);
            return;
        }
        setCandidateUsers(delegateTask, userTask, roleUserIds);
    }

    /**
     * 设置职位候选用户
     *
     * @param delegateTask
     * @param userTask
     * @param positionUserIds
     */
    private void setPositionCandidateUser(DelegateTask delegateTask, UserTask userTask, List<String> positionUserIds) {
        setCandidateUsers(delegateTask, userTask, positionUserIds);
    }

    /**
     * 设置候选人用户
     *
     * @param delegateTask
     * @param userTask
     * @param userIds
     */
    private void setCandidateUsers(DelegateTask delegateTask, UserTask userTask, List<String> userIds) {
        List<UserDTO> userDTOs = userBaseService.listByUserIds(userIds);
        if (userDTOs.isEmpty()) {
            throw new SearchException("下节点候选人不存在");
        }
        LOG.debug("设置候选人");
        List<String> existUserIds = ArrayListUtil.listBeanStringIdValue(userDTOs, "userId", UserDTO.class);
        clearUserIdsCurrentTaskAssignee(existUserIds, delegateTask.getAssignee());
        userTask.setCandidateUsers(existUserIds);
    }

    /**
     * 清除用户列表中当前任务代理人
     *
     * @param userIds
     * @param assignee
     */
    private void clearUserIdsCurrentTaskAssignee(List<String> userIds, String assignee) {
        for (String userId : userIds) {
            if (StringUtils.equals(userId, assignee)) {
                userIds.remove(userId);
                break;
            }
        }
    }

    private void assigneeCount(UserTask userTask, OaNodeAssigneePO oaNodeAssigneePO) {

    }

    // 1、nrOfInstances 该会签环节中总共有多少个实例 2、nrOfActiveInstances 当前活动的实例的数量，即还没有 完成的实例数量。 3、nrOfCompletedInstances 已经完成的实例的数量

    // 设置会签
//        // 获取多实例配置
//        MultiInstanceLoopCharacteristics characteristics = new MultiInstanceLoopCharacteristics();
//        // 设置集合变量，统一设置成users
//        characteristics.setInputDataItem(Constant.ACT_MUIT_LIST_NAME);
//        // 设置变量
//        characteristics.setElementVariable(Constant.ACT_MUIT_VAR_NAME);
//        // 设置为同时接收（false 表示不按顺序执行）
//        characteristics.setSequential(false);
//        // 设置条件（暂时处理成，全部会签完转下步）
//        characteristics.setCompletionCondition("${nrOfCompletedInstances==nrOfInstances}");
//
//        userTask.setLoopCharacteristics(characteristics);


    // 清空会签
    // 获取多实例配置
//        MultiInstanceLoopCharacteristics characteristics = userTask.getLoopCharacteristics();
//        if (characteristics != null) {
//            // 清空集合
//            characteristics.setInputDataItem("");
//            // 清空变量
//            characteristics.setElementVariable("");
//            // 设置为顺序接收（true 表示不按顺序执行）
//            characteristics.setSequential(true);
//            // 清空条件
//            characteristics.setCompletionCondition("");
//        }

}
